#region License Information
/* HeuristicLab
 * Copyright (C) 2002-2014 Heuristic and Evolutionary Algorithms Laboratory (HEAL)
 *
 * This file is part of HeuristicLab.
 *
 * HeuristicLab is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * HeuristicLab is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HeuristicLab. If not, see <http://www.gnu.org/licenses/>.
 */
#endregion

using System;
using System.Collections.Generic;
using System.Drawing;
using HeuristicLab.Common;
using HeuristicLab.Core;
using HeuristicLab.Persistence.Default.CompositeSerializers.Storable;

namespace HeuristicLab.Optimization.Networks {
  [Item("AlgorithmNode", "A node of an optimization network which contains a HeuristicLab algorithm.")]
  [StorableClass]
  public class AlgorithmNode : Node {
    [Storable]
    private IAlgorithm algorithm;
    public IAlgorithm Algorithm {
      get { return algorithm; }
      set {
        if (value != algorithm) {
          DeregisterAlgorithmEvents();
          algorithm = value;
          RegisterAlgorithmEvents();
          OnAlgorithmChanged();
        }
      }
    }

    [StorableConstructor]
    protected AlgorithmNode(bool deserializing) : base(deserializing) { }
    protected AlgorithmNode(AlgorithmNode original, Cloner cloner)
      : base(original, cloner) {
      algorithm = cloner.Clone(original.algorithm);
      RegisterPortsEvents();
      RegisterAlgorithmEvents();
    }
    public AlgorithmNode()
      : base("Node") {
      RegisterPortsEvents();
    }
    public AlgorithmNode(string name)
      : base(name) {
      RegisterPortsEvents();
    }
    public AlgorithmNode(string name, string description)
      : base(name, description) {
      RegisterPortsEvents();
    }

    [StorableHook(HookType.AfterDeserialization)]
    private void AfterDeserialization() {
      RegisterPortsEvents();
    }

    public override IDeepCloneable Clone(Cloner cloner) {
      return new AlgorithmNode(this, cloner);
    }

    private void UpdateParameter(IInputPort port) {
      if (algorithm != null) {
        foreach (var param in algorithm.Parameters) {
          IValueParameter p = param as IValueParameter;
          if ((p != null) && (p.Name == port.Name)) {
            p.Value = (IItem)port.Value;
          }
        }
      }
    }
    private void UpdateOutputPort(IResult result) {
      foreach (var port in Ports) {
        IOutputPort p = port as IOutputPort;
        if ((p != null) && (p.Name == result.Name)) {
          p.Value = result.Value;
        }
      }
    }

    public event EventHandler AlgorithmChanged;
    protected virtual void OnAlgorithmChanged() {
      var handler = AlgorithmChanged;
      if (handler != null) handler(this, EventArgs.Empty);
    }

    #region Ports Events
    private void RegisterPortsEvents() {
      Ports.ItemsAdded += Ports_ItemsAdded;
      Ports.ItemsRemoved += Ports_ItemsRemoved;
      Ports.ItemsReplaced += Ports_ItemsReplaced;
      Ports.CollectionReset += Ports_CollectionReset;
    }
    void Ports_ItemsAdded(object sender, Collections.CollectionItemsChangedEventArgs<IPort> e) {
      foreach (var p in e.Items)
        RegisterPortEvents(p);
    }
    void Ports_ItemsRemoved(object sender, Collections.CollectionItemsChangedEventArgs<IPort> e) {
      foreach (var p in e.Items)
        DeregisterPortEvents(p);
    }
    void Ports_ItemsReplaced(object sender, Collections.CollectionItemsChangedEventArgs<IPort> e) {
      foreach (var p in e.OldItems)
        DeregisterPortEvents(p);
      foreach (var p in e.Items)
        RegisterPortEvents(p);
    }
    void Ports_CollectionReset(object sender, Collections.CollectionItemsChangedEventArgs<IPort> e) {
      foreach (var p in e.OldItems)
        DeregisterPortEvents(p);
      foreach (var p in e.Items)
        RegisterPortEvents(p);
    }
    #endregion

    #region Port Events
    private void RegisterPortEvents(IPort port) {
      IInputPort p = port as IInputPort;
      if (p != null) {
        p.ValueChanged += InputPort_ValueChanged;
      }
    }
    private void DeregisterPortEvents(IPort port) {
      IInputPort p = port as IInputPort;
      if (p != null) {
        p.ValueChanged -= InputPort_ValueChanged;
      }
    }
    private void InputPort_ValueChanged(object sender, EventArgs e) {
      UpdateParameter((IInputPort)sender);
    }
    #endregion

    #region Algorithm Events
    private void RegisterAlgorithmEvents() {
      if (algorithm != null) {
        algorithm.Results.ItemsAdded += Results_ItemsAdded;
        algorithm.Results.ItemsRemoved += Results_ItemsRemoved;
        algorithm.Results.ItemsReplaced += Results_ItemsReplaced;
        algorithm.Results.CollectionReset += Results_CollectionReset;
      }
    }
    private void DeregisterAlgorithmEvents() {
      if (algorithm != null) {
        algorithm.Results.ItemsAdded -= Results_ItemsAdded;
        algorithm.Results.ItemsRemoved -= Results_ItemsRemoved;
        algorithm.Results.ItemsReplaced -= Results_ItemsReplaced;
        algorithm.Results.CollectionReset -= Results_CollectionReset;
      }
    }
    private void Results_ItemsAdded(object sender, Collections.CollectionItemsChangedEventArgs<IResult> e) {
      foreach (var r in e.Items)
        RegisterResultEvents(r);
    }
    private void Results_ItemsRemoved(object sender, Collections.CollectionItemsChangedEventArgs<IResult> e) {
      foreach (var r in e.Items)
        DeregisterResultEvents(r);
    }
    private void Results_ItemsReplaced(object sender, Collections.CollectionItemsChangedEventArgs<IResult> e) {
      foreach (var r in e.OldItems)
        DeregisterResultEvents(r);
      foreach (var r in e.Items)
        RegisterResultEvents(r);
    }
    private void Results_CollectionReset(object sender, Collections.CollectionItemsChangedEventArgs<IResult> e) {
      foreach (var r in e.OldItems)
        DeregisterResultEvents(r);
      foreach (var r in e.Items)
        RegisterResultEvents(r);
    }
    #endregion

    #region Result Events
    private void RegisterResultEvents(IResult result) {
      if (result != null) {
        result.ValueChanged += Result_ValueChanged;
      }
    }
    private void DeregisterResultEvents(IResult result) {
      if (result != null) {
        result.ValueChanged -= Result_ValueChanged;
      }
    }
    private void Result_ValueChanged(object sender, EventArgs e) {
      UpdateOutputPort((IResult)sender);
    }
    #endregion
  }
}
